from datetime import date
from decimal import Decimal

import netsgiro


def test_transmission_with_single_payment_request_transaction():
    transmission = netsgiro.Transmission(
        number='1703231',
        data_transmitter='01234567',
        data_recipient=netsgiro.NETS_ID,
    )
    assignment = transmission.add_assignment(
        service_code=netsgiro.ServiceCode.AVTALEGIRO,
        assignment_type=netsgiro.AssignmentType.TRANSACTIONS,
        number='0323001',
        account='15035382752',
    )
    assignment.add_payment_request(
        kid='000133700501645',
        due_date=date(2017, 4, 6),
        amount=Decimal('5244.63'),
        reference='ACME invoice #50164',
        payer_name='Wonderland',
        bank_notification=None,
    )
    assignment.add_payment_request(
        kid='001054300504897',
        due_date=date(2017, 4, 8),
        amount=Decimal('475.55'),
        reference='ACME invoice #50489',
        payer_name='Charlie',
        bank_notification='Foo bar',
    )

    records = list(transmission.to_records())

    assert len(records) == 10

    transmission_start = records[0]
    assignment_start = records[1]
    tx1_item1 = records[2]
    tx1_item2 = records[3]
    tx2_item1 = records[4]
    tx2_item2 = records[5]
    tx2_spec1 = records[6]
    tx2_spec2 = records[7]
    assignment_end = records[-2]
    transmission_end = records[-1]

    assert transmission_start.transmission_number == '1703231'
    assert transmission_start.data_transmitter == '01234567'
    assert transmission_start.data_recipient == netsgiro.NETS_ID
    assert transmission_end.nets_date == date(2017, 4, 6)

    assert assignment_start.assignment_number == '0323001'
    assert assignment_start.assignment_account == '15035382752'
    assert assignment_end.nets_date_earliest == date(2017, 4, 6)
    assert assignment_end.nets_date_latest == date(2017, 4, 8)

    assert tx1_item1.kid == '000133700501645'
    assert tx1_item1.amount == 524463
    assert tx1_item1.nets_date == date(2017, 4, 6)
    assert tx1_item2.reference == 'ACME invoice #50164'
    assert tx1_item2.payer_name == 'Wonderland'

    assert tx2_item1.kid == '001054300504897'
    assert tx2_item1.amount == 47555
    assert tx2_item1.nets_date == date(2017, 4, 8)
    assert tx2_item2.reference == 'ACME invoice #50489'
    assert tx2_item2.payer_name == 'Charlie'
    assert tx2_spec1.line_number == 1
    assert tx2_spec1.column_number == 1
    assert tx2_spec1.text == 'Foo bar                                 '
    assert tx2_spec2.line_number == 1
    assert tx2_spec2.column_number == 2
    assert tx2_spec2.text == '                                        '



"""The lower-level records API."""

import datetime
import re
from typing import Iterable, List, Optional, Sequence, Tuple, Union

import attr
from attr.validators import instance_of, optional

import netsgiro
from netsgiro.validators import str_of_length


__all__ = [
    'TransmissionStart',
    'TransmissionEnd',
    'AssignmentStart',
    'AssignmentEnd',
    'TransactionAmountItem1',
    'TransactionAmountItem2',
    'TransactionAmountItem3',
    'TransactionSpecification',
    'AvtaleGiroAgreement',
    'parse',
]


def to_service_code(
        value: Union[netsgiro.ServiceCode, int, str]) -> netsgiro.ServiceCode:
    return netsgiro.ServiceCode(int(value))


def to_assignment_type(
        value: Union[netsgiro.AssignmentType, int, str]
        ) -> netsgiro.AssignmentType:
    return netsgiro.AssignmentType(int(value))


def to_transaction_type(
        value: Union[netsgiro.TransactionType, int, str]
        ) -> netsgiro.TransactionType:
    return netsgiro.TransactionType(int(value))


def to_avtalegiro_registration_type(
        value: Union[netsgiro.AvtaleGiroRegistrationType, int, str]
        ) -> netsgiro.AvtaleGiroRegistrationType:
    return netsgiro.AvtaleGiroRegistrationType(int(value))


def to_date(value: Union[datetime.date, str, None]) -> Optional[datetime.date]:
    if isinstance(value, datetime.date):
        return value
    if value is None or value == '000000':
        return None
    return datetime.datetime.strptime(value, '%d%m%y').date()


def to_bool(value: Union[bool, str]) -> bool:
    if isinstance(value, bool):
        return value
    if value == 'J':
        return True
    elif value == 'N':
        return False
    else:
        raise ValueError("Expected 'J' or 'N', got {!r}".format(value))


def optional_int(value: Union[int, str, None]) -> Optional[int]:
    if value is None:
        return None
    else:
        return int(value)


def optional_str(value: Optional[str]) -> Optional[str]:
    return value and value.strip() or None


@attr.s
class Record:
    service_code = attr.ib(convert=to_service_code)

    _PATTERNS = []

    @classmethod
    def from_string(cls, line: str) -> 'Record':
        """Parse OCR string into a record object."""

        for pattern in cls._PATTERNS:
            matches = pattern.match(line)
            if matches is not None:
                return cls(**matches.groupdict())

        raise ValueError(
            '{!r} did not match {} record formats'
            .format(line, cls.__name__))

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        raise NotImplementedError


@attr.s
class TransmissionStart(Record):
    """TransmissionStart is the first record in every OCR file.

    A file can only contain a single transmission.

    Each transmission can contain any number of assignments.
    """

    transmission_number = attr.ib(validator=str_of_length(7))
    data_transmitter = attr.ib(validator=str_of_length(8))
    data_recipient = attr.ib(validator=str_of_length(8))

    RECORD_TYPE = netsgiro.RecordType.TRANSMISSION_START
    _PATTERNS = [
        re.compile(r'''
            ^
            NY      # Format code
            (?P<service_code>00)
            00      # Transmission type, always 00
            10      # Record type

            (?P<data_transmitter>\d{8})
            (?P<transmission_number>\d{7})
            (?P<data_recipient>\d{8})

            0{49}   # Padding
            $
        ''', re.VERBOSE),
    ]

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        return (
            'NY000010'
            '{self.data_transmitter:8}'
            '{self.transmission_number:7}'
            '{self.data_recipient:8}'
            + ('0' * 49)
        ).format(self=self)


@attr.s
class TransmissionEnd(Record):
    """TransmissionEnd is the first record in every OCR file."""

    num_transactions = attr.ib(convert=int)
    num_records = attr.ib(convert=int)
    total_amount = attr.ib(convert=int)
    nets_date = attr.ib(convert=to_date)

    RECORD_TYPE = netsgiro.RecordType.TRANSMISSION_END
    _PATTERNS = [
        re.compile(r'''
            ^
            NY      # Format code
            (?P<service_code>00)
            00      # Transmission type, always 00
            89      # Record type

            (?P<num_transactions>\d{8})
            (?P<num_records>\d{8})
            (?P<total_amount>\d{17})
            (?P<nets_date>\d{6})

            0{33}   # Filler
            $
        ''', re.VERBOSE),
    ]

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        return (
            'NY000089'
            '{self.num_transactions:08d}'
            '{self.num_records:08d}'
            '{self.total_amount:017d}'
            '{self.nets_date:%d%m%y}'
            + ('0' * 33)
        ).format(self=self)


@attr.s
class AssignmentStart(Record):
    """AssignmentStart is the first record of an assignment.

    Each assignment can contain any number of transactions.
    """

    assignment_type = attr.ib(convert=to_assignment_type)
    assignment_number = attr.ib(validator=str_of_length(7))
    assignment_account = attr.ib(validator=str_of_length(11))

    # Only for assignment_type == AssignmentType.TRANSACTIONS
    agreement_id = attr.ib(default=None, validator=optional(str_of_length(9)))

    RECORD_TYPE = netsgiro.RecordType.ASSIGNMENT_START
    _PATTERNS = [
        re.compile(r'''
            ^
            NY      # Format code
            (?P<service_code>(09|21))
            (?P<assignment_type>00)
            20      # Record type

            (?P<agreement_id>\d{9})
            (?P<assignment_number>\d{7})
            (?P<assignment_account>\d{11})

            0{45}   # Filler
            $
        ''', re.VERBOSE),
        re.compile(r'''
            ^
            NY      # Format code
            (?P<service_code>21)
            (?P<assignment_type>24)
            20      # Record type

            0{9}    # Filler

            (?P<assignment_number>\d{7})
            (?P<assignment_account>\d{11})

            0{45}   # Filler
            $
        ''', re.VERBOSE),
        re.compile(r'''
            ^
            NY      # Format code
            (?P<service_code>21)
            (?P<assignment_type>36)
            20      # Record type

            0{9}    # Filler

            (?P<assignment_number>\d{7})
            (?P<assignment_account>\d{11})

            0{45}   # Filler
            $
        ''', re.VERBOSE),
    ]

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        return (
            'NY'
            '{self.service_code:02d}'
            '{self.assignment_type:02d}'
            '20'
            + (self.agreement_id and '{self.agreement_id:9}' or ('0' * 9)) +
            '{self.assignment_number:7}'
            '{self.assignment_account:11}'
            + ('0' * 45)
        ).format(self=self)


@attr.s
class AssignmentEnd(Record):
    """AssignmentEnd is the last record of an assignment."""

    assignment_type = attr.ib(convert=to_assignment_type)
    num_transactions = attr.ib(convert=int)
    num_records = attr.ib(convert=int)

    # Only for transactions and cancellations
    total_amount = attr.ib(default=None, convert=optional_int)
    nets_date_1 = attr.ib(default=None, convert=to_date)
    nets_date_2 = attr.ib(default=None, convert=to_date)
    nets_date_3 = attr.ib(default=None, convert=to_date)

    RECORD_TYPE = netsgiro.RecordType.ASSIGNMENT_END
    _PATTERNS = [
        re.compile(r'''
            ^
            NY      # Format code
            (?P<service_code>(09|21))
            (?P<assignment_type>00)     # Transactions / payment requests
            88      # Record type

            (?P<num_transactions>\d{8})
            (?P<num_records>\d{8})
            (?P<total_amount>\d{17})
            (?P<nets_date_1>\d{6})
            (?P<nets_date_2>\d{6})
            (?P<nets_date_3>\d{6})

            0{21}   # Filler
            $
        ''', re.VERBOSE),
        re.compile(r'''
            ^
            NY      # Format code
            (?P<service_code>21)
            (?P<assignment_type>24)     # AvtaleGiro agreements
            88      # Record type

            (?P<num_transactions>\d{8})
            (?P<num_records>\d{8})

            0{56}   # Filler
            $
        ''', re.VERBOSE),
        re.compile(r'''
            ^
            NY      # Format code
            (?P<service_code>21)
            (?P<assignment_type>36)     # AvtaleGiro cancellations
            88      # Record type

            (?P<num_transactions>\d{8})
            (?P<num_records>\d{8})
            (?P<total_amount>\d{17})
            (?P<nets_date_1>\d{6})
            (?P<nets_date_2>\d{6})

            0{27}   # Filler
            $
        ''', re.VERBOSE),
    ]

    @property
    def nets_date(self):
        """Nets' processing date.

        Only used for OCR Giro.
        """
        if self.service_code == netsgiro.ServiceCode.OCR_GIRO:
            return self.nets_date_1
        else:
            return None

    @property
    def nets_date_earliest(self):
        """Earliest date from the contained transactions."""

        if self.service_code == netsgiro.ServiceCode.OCR_GIRO:
            return self.nets_date_2
        elif self.service_code == netsgiro.ServiceCode.AVTALEGIRO:
            return self.nets_date_1
        else:
            raise ValueError(
                'Unhandled service code: {}'.format(self.service_code))

    @property
    def nets_date_latest(self):
        """Latest date from the contained transactions."""

        if self.service_code == netsgiro.ServiceCode.OCR_GIRO:
            return self.nets_date_3
        elif self.service_code == netsgiro.ServiceCode.AVTALEGIRO:
            return self.nets_date_2
        else:
            raise ValueError(
                'Unhandled service code: {}'.format(self.service_code))

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        return (
            'NY'
            '{self.service_code:02d}'
            '{self.assignment_type:02d}'
            '88'
            '{self.num_transactions:08d}'
            '{self.num_records:08d}'
            + (self.total_amount and '{self.total_amount:017d}' or ('0' * 17))
            + (self.nets_date_1 and '{self.nets_date_1:%d%m%y}' or ('0' * 6))
            + (self.nets_date_2 and '{self.nets_date_2:%d%m%y}' or ('0' * 6))
            + (self.nets_date_3 and '{self.nets_date_3:%d%m%y}' or ('0' * 6))
            + ('0' * 21)
        ).format(self=self)


@attr.s
class TransactionRecord(Record):
    transaction_type = attr.ib(convert=to_transaction_type)
    transaction_number = attr.ib(convert=int)


@attr.s
class TransactionAmountItem1(TransactionRecord):
    """TransactionAmountItem1 is the first record of a transaction.

    The record is used both for AvtaleGiro and for OCR Giro.
    """

    nets_date = attr.ib(convert=to_date)
    amount = attr.ib(convert=int)
    kid = attr.ib(convert=optional_str)

    # Only OCR Giro
    centre_id = attr.ib(default=None, validator=optional(str_of_length(2)))
    day_code = attr.ib(default=None, convert=optional_int)
    partial_settlement_number = attr.ib(default=None, convert=optional_int)
    partial_settlement_serial_number = attr.ib(
        default=None, validator=optional(str_of_length(5)))
    sign = attr.ib(default=None, validator=optional(str_of_length(1)))

    RECORD_TYPE = netsgiro.RecordType.TRANSACTION_AMOUNT_ITEM_1
    _PATTERNS = [
        re.compile(r'''
            ^
            NY      # Format code
            (?P<service_code>09)
            (?P<transaction_type>\d{2})  # 10-21
            30      # Record type

            (?P<transaction_number>\d{7})
            (?P<nets_date>\d{6})

            (?P<centre_id>\d{2})
            (?P<day_code>\d{2})
            (?P<partial_settlement_number>\d{1})
            (?P<partial_settlement_serial_number>\d{5})
            (?P<sign>[-0]{1})

            (?P<amount>\d{17})
            (?P<kid>[\d ]{25})

            0{6}    # Filler
            $
        ''', re.VERBOSE),
        re.compile(r'''
            ^
            NY      # Format code
            (?P<service_code>21)
            (?P<transaction_type>\d{2})  # 02, 21, or 93
            30      # Record type

            (?P<transaction_number>\d{7})
            (?P<nets_date>\d{6})

            [ ]{11} # Filler

            (?P<amount>\d{17})
            (?P<kid>[\d ]{25})

            0{6}    # Filler
            $
        ''', re.VERBOSE),
    ]

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        if self.service_code == netsgiro.ServiceCode.OCR_GIRO:
            ocr_giro_fields = (
                '{self.centre_id:2}'
                '{self.day_code:02d}'
                '{self.partial_settlement_number:01d}'
                '{self.partial_settlement_serial_number:5}'
                '{self.sign:1}'
            ).format(self=self)
        else:
            ocr_giro_fields = ' ' * 11

        return (
            'NY'
            '{self.service_code:02d}'
            '{self.transaction_type:02d}'
            '30'
            '{self.transaction_number:07d}'
            '{self.nets_date:%d%m%y}'
            + ocr_giro_fields +
            '{self.amount:017d}'
            '{self.kid:>25}'
            + ('0' * 6)
        ).format(self=self)


@attr.s
class TransactionAmountItem2(TransactionRecord):
    """TransactionAmountItem2 is the second record of a transaction.

    The record is used both for AvtaleGiro and for OCR Giro.
    """

    reference = attr.ib(convert=optional_str)

    # Only OCR Giro
    form_number = attr.ib(default=None, validator=optional(str_of_length(10)))
    bank_date = attr.ib(default=None, convert=to_date)
    debit_account = attr.ib(
        default=None, validator=optional(str_of_length(11)))
    # XXX In use in OCR Giro "from giro debited account" transactions in test
    # data, but documented as a filler field.
    _filler = attr.ib(default=None)

    # Only AvtaleGiro
    payer_name = attr.ib(default=None, convert=optional_str)

    RECORD_TYPE = netsgiro.RecordType.TRANSACTION_AMOUNT_ITEM_2
    _PATTERNS = [
        re.compile(r'''
            ^
            NY      # Format code
            (?P<service_code>09)
            (?P<transaction_type>\d{2})  # 10-21
            31      # Record type

            (?P<transaction_number>\d{7})
            (?P<form_number>\d{10})
            (?P<reference>\d{9})

            (?P<filler>.{7})  # XXX Documented as filler, in use in test data

            (?P<bank_date>\d{6})
            (?P<debit_account>\d{11})

            0{22}    # Filler
            $
        ''', re.VERBOSE),
        re.compile(r'''
            ^
            NY      # Format code
            (?P<service_code>21)
            (?P<transaction_type>\d{2})  # 02, 21, or 93
            31      # Record type

            (?P<transaction_number>\d{7})
            (?P<payer_name>.{10})

            [ ]{25} # Filler

            (?P<reference>.{25})

            0{5}    # Filler
            $
        ''', re.VERBOSE),
    ]

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        common_fields = (
            'NY'
            '{self.service_code:02d}'
            '{self.transaction_type:02d}'
            '31'
            '{self.transaction_number:07d}'
        ).format(self=self)

        if self.service_code == netsgiro.ServiceCode.OCR_GIRO:
            service_fields = (
                '{self.form_number:10}'
                + (self.reference and '{self.reference:9}' or (' ' * 9))
                + (self._filler and '{self._filler:7}' or ('0' * 7))
                + (self.bank_date and '{self.bank_date:%d%m%y}' or '0' * 6) +
                '{self.debit_account:11}'
                + ('0' * 22)
            ).format(self=self)
        elif self.service_code == netsgiro.ServiceCode.AVTALEGIRO:
            service_fields = (
                '{:10}'.format(self.payer_name[:10])
                + (' ' * 25)
                + (self.reference and '{self.reference:25}' or (' ' * 25))
                + ('0' * 5)
            ).format(self=self)
        else:
            service_fields = ' ' * 35

        return common_fields + service_fields


@attr.s
class TransactionAmountItem3(TransactionRecord):
    """TransactionAmountItem3 is the third record of a transaction.

    The record is only used for some OCR Giro transaction types.
    """

    text = attr.ib(convert=optional_str)

    RECORD_TYPE = netsgiro.RecordType.TRANSACTION_AMOUNT_ITEM_3
    _PATTERNS = [
        re.compile(r'''
            ^
            NY      # Format code
            (?P<service_code>09)
            (?P<transaction_type>\d{2})  # 20-21
            32      # Record type

            (?P<transaction_number>\d{7})
            (?P<text>.{40})

            0{25}    # Filler
            $
        ''', re.VERBOSE),
    ]

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        return (
            'NY09'
            '{self.transaction_type:02d}'
            '32'
            '{self.transaction_number:07d}'
            '{self.text:40}'
            + ('0' * 25)
        ).format(self=self)


@attr.s
class TransactionSpecification(TransactionRecord):
    """TransactionSpecification is used for AvtaleGiro transactions.

    The record is only used when bank notification is used to notify the payer.

    Each record contains half of an 80 char long line of text and can be
    repeated up to 84 times for a single transaction for a total of 42 lines of
    specification text.
    """

    line_number = attr.ib(convert=int)
    column_number = attr.ib(convert=int)
    text = attr.ib(validator=instance_of(str))

    RECORD_TYPE = netsgiro.RecordType.TRANSACTION_SPECIFICATION
    _PATTERNS = [
        re.compile(r'''
            ^
            NY      # Format code
            (?P<service_code>21)
            (?P<transaction_type>21)
            49      # Record type

            (?P<transaction_number>\d{7})
            4       # Payment notification
            (?P<line_number>\d{3})
            (?P<column_number>\d{1})
            (?P<text>.{40})

            0{20}    # Filler
            $
        ''', re.VERBOSE),
    ]

    _MAX_LINES = 42
    _MAX_LINE_LENGTH = 80
    _MAX_COLUMNS = 2
    _MAX_RECORDS = _MAX_LINES * _MAX_COLUMNS

    @classmethod
    def from_text(
            cls, *,
            service_code, transaction_type, transaction_number, text
            ) -> Iterable['TransactionSpecification']:
        """Create a sequence of specification records from a text string."""

        for line, column, text in (
                cls._split_text_to_lines_and_columns(text)):
            yield cls(
                service_code=service_code,
                transaction_type=transaction_type,
                transaction_number=transaction_number,

                line_number=line,
                column_number=column,
                text=text,
            )

    @classmethod
    def _split_text_to_lines_and_columns(
            cls, text) -> Iterable[Tuple[int, int, str]]:
        lines = text.splitlines()

        if len(lines) > cls._MAX_LINES:
            raise ValueError(
                'Max {} specification lines allowed, got {}'
                .format(cls._MAX_LINES, len(lines)))

        for line_number, line_text in enumerate(lines, 1):
            if len(line_text) > cls._MAX_LINE_LENGTH:
                raise ValueError(
                    'Specification lines must be max {} chars long, '
                    'got {}: {!r}'
                    .format(cls._MAX_LINE_LENGTH, len(line_text), line_text))

            yield (line_number, 1, '{:40}'.format(line_text[0:40]))
            yield (line_number, 2, '{:40}'.format(line_text[40:80]))

    @classmethod
    def to_text(cls, records: Sequence['TransactionSpecification']) -> str:
        """Get a text string from a sequence of specification records."""

        if len(records) > cls._MAX_RECORDS:
            raise ValueError(
                'Max {} specification records allowed, got {}'
                .format(cls._MAX_RECORDS, len(records)))

        tuples = sorted([
            (r.line_number, r.column_number, r)
            for r in records
        ])

        text = ''
        for _, column, specification in tuples:
            text += specification.text
            if column == cls._MAX_COLUMNS:
                text += '\n'

        return text

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        return (
            'NY212149'
            '{self.transaction_number:07d}'
            '4'
            '{self.line_number:03d}'
            '{self.column_number:01d}'
            '{self.text:40}'
            + ('0' * 20)
        ).format(self=self)


@attr.s
class AvtaleGiroAgreement(TransactionRecord):
    """AvtaleGiroAgreement is used by Nets to notify about agreement changes.

    This includes new or deleted agreements, as well as updates to the payer's
    notification preferences.
    """

    registration_type = attr.ib(convert=to_avtalegiro_registration_type)
    kid = attr.ib(convert=optional_str)
    notify = attr.ib(convert=to_bool)

    RECORD_TYPE = netsgiro.RecordType.TRANSACTION_AGREEMENTS
    _PATTERNS = [
        re.compile(r'''
            ^
            NY      # Format code
            (?P<service_code>21)
            (?P<transaction_type>94)
            70      # Record type

            (?P<transaction_number>\d{7})
            (?P<registration_type>\d{1})
            (?P<kid>[\d ]{25})
            (?P<notify>[JN]{1})

            0{38}   # Filler
            $
        ''', re.VERBOSE),
    ]

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        return (
            'NY219470'
            '{self.transaction_number:07d}'
            '{self.registration_type:01d}'
            '{self.kid:>25}'
            + (self.notify and 'J' or 'N')
            + ('0' * 38)
        ).format(self=self)


def parse(data: str) -> List[Record]:
    """Parse an OCR file into a list of record objects."""

    def all_subclasses(cls):
        return cls.__subclasses__() + [
            subsubcls
            for subcls in cls.__subclasses__()
            for subsubcls in all_subclasses(subcls)]

    record_classes = {
        cls.RECORD_TYPE: cls
        for cls in all_subclasses(Record)
        if hasattr(cls, 'RECORD_TYPE')
    }

    results = []

    for line in data.strip().splitlines():
        if len(line) != 80:
            raise ValueError('All lines must be exactly 80 chars long')

        record_type_str = line[6:8]
        if not record_type_str.isnumeric():
            raise ValueError(
                'Record type must be numeric, got {!r}'
                .format(record_type_str))

        record_type = netsgiro.RecordType(int(record_type_str))
        record_cls = record_classes[record_type]

        results.append(record_cls.from_string(line))

    return results



import os
import sys

import sphinx_rtd_theme

sys.path.insert(0, os.path.abspath('..'))

import netsgiro  # noqa


# -- General configuration ------------------------------------------------

# needs_sphinx = '1.0'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.viewcode',
]

templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'

project = 'netsgiro'
copyright = '2017, Otovo AS'
author = 'Otovo AS'

release = netsgiro.__version__
version = '.'.join(release.split('.')[:2])

language = None
exclude_patterns = ['_build']

pygments_style = 'sphinx'

todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

html_theme = 'sphinx_rtd_theme'
html_theme_options = {}
html_theme_options['canonical_url'] = (
    'https://netsgiro.readthedocs.io/en/latest/')
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_static_path = ['_static']


# -- Options for HTMLHelp output ------------------------------------------

htmlhelp_basename = 'netsgiro'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    'papersize': 'a4paper',
    # 'pointsize': '10pt',
    # 'preamble': '',
    # 'figure_align': 'htbp',
}

latex_documents = [
    (
        master_doc,
        'netsgiro.tex',
        'netsgiro documentation',
        'Otovo AS',
        'manual',
    ),
]


# -- Options for doctest extension ----------------------------------------

autodoc_member_order = 'groupwise'


# -- Options for doctest builder ------------------------------------------

doctest_path = [
    os.path.abspath('..'),
]

doctest_global_setup = """
from pprint import pprint
"""

doctest_test_doctest_blocks = 'default'



from pathlib import Path

import pytest


TEST_DIR = Path(__file__).parent


@pytest.fixture
def agreements_data():
    filepath = TEST_DIR / 'data' / 'avtalegiro_agreements.txt'
    with filepath.open('r', encoding='iso-8859-1') as fh:
        return fh.read()


@pytest.fixture
def payment_request_data():
    filepath = TEST_DIR / 'data' / 'avtalegiro_payment_request.txt'
    with filepath.open('r', encoding='iso-8859-1') as fh:
        return fh.read()


@pytest.fixture
def ocr_giro_transactions_data():
    filepath = TEST_DIR / 'data' / 'ocr_giro_transactions.txt'
    with filepath.open('r', encoding='iso-8859-1') as fh:
        return fh.read()



import netsgiro


def test_write_agreements(agreements_data):
    transmission = netsgiro.parse(agreements_data)

    result = transmission.to_ocr()

    assert result == agreements_data.strip()


def test_write_payment_request(payment_request_data):
    transmission = netsgiro.parse(payment_request_data)

    result = transmission.to_ocr()

    assert result == payment_request_data.strip()


def test_write_ocr_giro_transactions(ocr_giro_transactions_data):
    transmission = netsgiro.parse(ocr_giro_transactions_data)

    result = transmission.to_ocr()

    assert result == ocr_giro_transactions_data.strip()



from datetime import date

import pytest

import netsgiro
import netsgiro.records


def test_parse_with_too_short_lines_fails():
    with pytest.raises(ValueError) as exc_info:
        netsgiro.records.parse(
            'NY0000\n'
            'NY0000\n'
        )

    assert 'exactly 80 chars long' in str(exc_info)


def test_parse_with_nonnumeric_record_type():
    with pytest.raises(ValueError) as exc_info:
        netsgiro.records.parse('NY0000AA' + '0' * 72)

    assert "Record type must be numeric, got 'AA'" in str(exc_info)


def test_parse_with_unknown_record_type():
    with pytest.raises(ValueError) as exc_info:
        netsgiro.records.parse('NY000099' + '0' * 72)

    assert '99 is not a valid RecordType' in str(exc_info)


def test_parse_avtalegiro_agreements(agreements_data):
    result = netsgiro.records.parse(agreements_data)

    assert len(result) == 20

    transmission_start = result[0]
    assignment_start = result[1]
    agreement_1 = result[2]
    agreement_2 = result[3]
    assignment_end = result[-2]
    transmission_end = result[-1]

    assert isinstance(transmission_start, netsgiro.records.TransmissionStart)
    assert transmission_start.transmission_number == '1091949'
    assert transmission_start.data_transmitter == netsgiro.NETS_ID

    assert isinstance(assignment_start, netsgiro.records.AssignmentStart)
    assert assignment_start.assignment_account == '99991042764'

    assert isinstance(agreement_1, netsgiro.records.AvtaleGiroAgreement)
    assert agreement_1.registration_type == (
        netsgiro.AvtaleGiroRegistrationType.NEW_OR_UPDATED_AGREEMENT)
    assert agreement_1.kid == '000112000507155'
    assert agreement_1.notify is True

    assert isinstance(agreement_2, netsgiro.records.AvtaleGiroAgreement)
    assert agreement_2.registration_type == (
        netsgiro.AvtaleGiroRegistrationType.NEW_OR_UPDATED_AGREEMENT)
    assert agreement_2.kid == '001006300507304'
    assert agreement_2.notify is False

    assert isinstance(assignment_end, netsgiro.records.AssignmentEnd)
    assert assignment_end.num_transactions == 16
    assert assignment_end.num_records == 18
    assert assignment_end.total_amount is None
    assert assignment_end.nets_date_earliest is None
    assert assignment_end.nets_date_latest is None

    assert isinstance(transmission_end, netsgiro.records.TransmissionEnd)
    assert transmission_end.num_transactions == 16
    assert transmission_end.num_records == 20
    assert transmission_end.total_amount == 0
    assert transmission_end.nets_date == date(2017, 4, 19)


def test_parse_avtalegiro_payment_requests(payment_request_data):
    result = netsgiro.records.parse(payment_request_data)

    assert len(result) == 22

    transmission_start = result[0]
    assignment_start = result[1]
    transaction_amount_1 = result[2]
    transaction_amount_2 = result[3]
    transaction_spec_1 = result[4]
    transaction_spec_2 = result[5]
    assignment_end = result[-2]
    transmission_end = result[-1]

    assert isinstance(transmission_start, netsgiro.records.TransmissionStart)
    assert transmission_start.transmission_number == '1000081'
    assert transmission_start.data_recipient == netsgiro.NETS_ID

    assert isinstance(assignment_start, netsgiro.records.AssignmentStart)
    assert assignment_start.assignment_account == '88888888888'

    assert isinstance(
        transaction_amount_1, netsgiro.records.TransactionAmountItem1)
    assert transaction_amount_1.nets_date == date(2004, 6, 17)
    assert transaction_amount_1.amount == 100
    assert transaction_amount_1.kid == '008000011688373'

    assert isinstance(
        transaction_amount_2, netsgiro.records.TransactionAmountItem2)
    assert transaction_amount_2.payer_name == 'NAVN'
    assert transaction_amount_2.reference is None

    assert isinstance(
        transaction_spec_1, netsgiro.records.TransactionSpecification)
    assert transaction_spec_1.text == (
        ' Gjelder Faktura: 168837  Dato: 19/03/04')

    assert isinstance(
        transaction_spec_2, netsgiro.records.TransactionSpecification)
    assert transaction_spec_2.text == (
        '                  ForfallsDato: 17/06/04')

    assert isinstance(assignment_end, netsgiro.records.AssignmentEnd)
    assert assignment_end.num_transactions == 6
    assert assignment_end.num_records == 20
    assert assignment_end.total_amount == 600
    assert assignment_end.nets_date_earliest == date(2004, 6, 17)
    assert assignment_end.nets_date_latest == date(2004, 6, 17)

    assert isinstance(transmission_end, netsgiro.records.TransmissionEnd)
    assert transmission_end.num_transactions == 6
    assert transmission_end.num_records == 22
    assert transmission_end.total_amount == 600
    assert transmission_end.nets_date == date(2004, 6, 17)


def test_parse_ocr_giro_transactions(ocr_giro_transactions_data):
    result = netsgiro.records.parse(ocr_giro_transactions_data)

    assert len(result) == 45

    transmission_start = result[0]
    assignment_start = result[1]
    transaction_amount_1 = result[2]
    transaction_amount_2 = result[3]
    transaction_amount_3 = result[4]
    assignment_end = result[-2]
    transmission_end = result[-1]

    assert isinstance(
        transmission_start, netsgiro.records.TransmissionStart)
    assert transmission_start.transmission_number == '0170031'
    assert transmission_start.data_recipient == '00010200'

    assert isinstance(
        assignment_start, netsgiro.records.AssignmentStart)
    assert assignment_start.assignment_account == '99991042764'

    assert isinstance(
        transaction_amount_1, netsgiro.records.TransactionAmountItem1)
    assert transaction_amount_1.nets_date == date(1992, 1, 20)
    assert transaction_amount_1.amount == 102000
    assert transaction_amount_1.kid == '0000531'

    assert isinstance(
        transaction_amount_2, netsgiro.records.TransactionAmountItem2)
    assert transaction_amount_2.payer_name is None
    assert transaction_amount_2.form_number == '9636827194'
    assert transaction_amount_2.reference == '099038562'
    assert transaction_amount_2.bank_date == date(1992, 1, 16)
    assert transaction_amount_2.debit_account == '99990512341'

    assert isinstance(
        transaction_amount_3, netsgiro.records.TransactionAmountItem3)
    assert transaction_amount_3.text == 'Foo bar baz'

    assert isinstance(assignment_end, netsgiro.records.AssignmentEnd)
    assert assignment_end.num_transactions == 20
    assert assignment_end.num_records == 43
    assert assignment_end.total_amount == 5144900
    assert assignment_end.nets_date == date(1992, 1, 20)
    assert assignment_end.nets_date_earliest == date(1992, 1, 20)
    assert assignment_end.nets_date_latest == date(1992, 1, 20)

    assert isinstance(transmission_end, netsgiro.records.TransmissionEnd)
    assert transmission_end.num_transactions == 20
    assert transmission_end.num_records == 45
    assert transmission_end.total_amount == 5144900
    assert transmission_end.nets_date == date(1992, 1, 20)



"""Custom validators for :mod:`attrs`."""

from attr.validators import instance_of


def str_of_length(length):
    """Validate that the value is a string of the given length."""

    def validator(instance, attribute, value):
        instance_of(str)(instance, attribute, value)
        if len(value) != length:
            raise ValueError(
                '{0.name} must be exactly {1} chars, got {2!r}'
                .format(attribute, length, value))
    return validator



__version__ = '1.0.0'


from netsgiro.constants import *  # noqa: Rexport
from netsgiro.enums import *  # noqa: Reexport
from netsgiro.objects import *  # noqa: Reexport


from netsgiro import constants, enums, objects

__all__ = (
    constants.__all__ +
    enums.__all__ +
    objects.__all__
)



from datetime import date

import pytest

import netsgiro


@pytest.fixture
def transmission():
    return netsgiro.Transmission(
        number='0000001',
        data_transmitter='12341234',
        data_recipient=netsgiro.NETS_ID,
        date=date(2004, 6, 17),
    )


def test_transmission_from_zero_records_fails():
    with pytest.raises(ValueError) as exc_info:
        netsgiro.Transmission.from_records([])

    assert 'At least 2 records required, got 0' in str(exc_info)


def test_assignment_from_zero_records_fails():
    with pytest.raises(ValueError) as exc_info:
        netsgiro.Assignment.from_records([])

    assert 'At least 2 records required, got 0' in str(exc_info)



from datetime import date

import pytest

import netsgiro
import netsgiro.records


def test_transmission_start():
    record = netsgiro.records.TransmissionStart.from_string(
        'NY00001055555555100008100008080000000000'
        '0000000000000000000000000000000000000000'
    )

    assert record.service_code == netsgiro.ServiceCode.NONE
    assert record.RECORD_TYPE == netsgiro.RecordType.TRANSMISSION_START

    assert record.data_transmitter == '55555555'
    assert record.transmission_number == '1000081'
    assert record.data_recipient == '00008080'


def test_transmission_start_fails_when_invalid_format():
    line = 'XX' + ('0' * 78)

    with pytest.raises(ValueError) as exc_info:
        netsgiro.records.TransmissionStart.from_string(line)

    assert (
        '{!r} did not match TransmissionStart record format'.format(line)
        in str(exc_info)
    )


def test_transmission_end():
    record = netsgiro.records.TransmissionEnd.from_string(
        'NY00008900000006000000220000000000000060'
        '0170604000000000000000000000000000000000'
    )

    assert record.service_code == netsgiro.ServiceCode.NONE
    assert record.RECORD_TYPE == netsgiro.RecordType.TRANSMISSION_END

    assert record.num_transactions == 6
    assert record.num_records == 22
    assert record.total_amount == 600
    assert record.nets_date == date(2004, 6, 17)


def test_assignment_start_for_avtalegiro_payment_requests():
    record = netsgiro.records.AssignmentStart.from_string(
        'NY21002000000000040000868888888888800000'
        '0000000000000000000000000000000000000000'
    )

    assert record.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.ASSIGNMENT_START

    assert record.assignment_type == netsgiro.AssignmentType.TRANSACTIONS

    assert record.agreement_id == '000000000'
    assert record.assignment_number == '4000086'
    assert record.assignment_account == '88888888888'


def test_assignment_start_for_avtalegiro_agreements():
    record = netsgiro.records.AssignmentStart.from_string(
        'NY21242000000000040000868888888888800000'
        '0000000000000000000000000000000000000000'
    )

    assert record.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.ASSIGNMENT_START

    assert record.assignment_type == (
        netsgiro.AssignmentType.AVTALEGIRO_AGREEMENTS)

    assert record.agreement_id is None
    assert record.assignment_number == '4000086'
    assert record.assignment_account == '88888888888'


def test_assignment_start_for_avtalegiro_cancellation():
    record = netsgiro.records.AssignmentStart.from_string(
        'NY21362000000000040000868888888888800000'
        '0000000000000000000000000000000000000000'
    )

    assert record.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.ASSIGNMENT_START

    assert record.assignment_type == (
        netsgiro.AssignmentType.AVTALEGIRO_CANCELLATIONS)

    assert record.agreement_id is None
    assert record.assignment_number == '4000086'
    assert record.assignment_account == '88888888888'


def test_assignment_start_for_ocr_giro_transactions():
    record = netsgiro.records.AssignmentStart.from_string(
        'NY09002000100856600000029999104276400000'
        '0000000000000000000000000000000000000000'
    )

    assert record.service_code == netsgiro.ServiceCode.OCR_GIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.ASSIGNMENT_START

    assert record.assignment_type == netsgiro.AssignmentType.TRANSACTIONS

    assert record.agreement_id == '001008566'
    assert record.assignment_number == '0000002'
    assert record.assignment_account == '99991042764'


def test_assignment_end_for_avtalegiro_payment_requests():
    record = netsgiro.records.AssignmentEnd.from_string(
        'NY21008800000006000000200000000000000060'
        '0170604170604000000000000000000000000000'
    )

    assert record.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.ASSIGNMENT_END

    assert record.assignment_type == netsgiro.AssignmentType.TRANSACTIONS

    assert record.num_transactions == 6
    assert record.num_records == 20
    assert record.total_amount == 600
    assert record.nets_date_earliest == date(2004, 6, 17)
    assert record.nets_date_latest == date(2004, 6, 17)


def test_assignment_end_for_avtalegiro_agreements():
    record = netsgiro.records.AssignmentEnd.from_string(
        'NY21248800000006000000200000000000000000'
        '0000000000000000000000000000000000000000'
    )

    assert record.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.ASSIGNMENT_END

    assert record.assignment_type == (
        netsgiro.AssignmentType.AVTALEGIRO_AGREEMENTS)

    assert record.num_transactions == 6
    assert record.num_records == 20
    assert record.total_amount is None
    assert record.nets_date_earliest is None
    assert record.nets_date_latest is None


def test_assignment_end_for_avtalegiro_cancellations():
    record = netsgiro.records.AssignmentEnd.from_string(
        'NY21368800000006000000200000000000000060'
        '0170604170604000000000000000000000000000'
    )

    assert record.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.ASSIGNMENT_END

    assert record.assignment_type == (
        netsgiro.AssignmentType.AVTALEGIRO_CANCELLATIONS)

    assert record.num_transactions == 6
    assert record.num_records == 20
    assert record.total_amount == 600
    assert record.nets_date_latest == date(2004, 6, 17)
    assert record.nets_date_earliest == date(2004, 6, 17)


def test_assignment_end_for_ocr_giro_transactions():
    record = netsgiro.records.AssignmentEnd.from_string(
        'NY09008800000020000000420000000000514490'
        '0200192200192200192000000000000000000000'
    )

    assert record.service_code == netsgiro.ServiceCode.OCR_GIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.ASSIGNMENT_END

    assert record.assignment_type == netsgiro.AssignmentType.TRANSACTIONS

    assert record.num_transactions == 20
    assert record.num_records == 42
    assert record.total_amount == 5144900
    assert record.nets_date == date(1992, 1, 20)
    assert record.nets_date_earliest == date(1992, 1, 20)
    assert record.nets_date_latest == date(1992, 1, 20)


def test_transaction_amount_item_1_for_avtalegiro_payment_request():
    record = netsgiro.records.TransactionAmountItem1.from_string(
        'NY2121300000001170604           00000000'
        '000000100          008000011688373000000'
    )

    assert record.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.TRANSACTION_AMOUNT_ITEM_1

    assert record.transaction_type == (
        netsgiro.TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION)
    assert record.transaction_number == 1

    assert record.nets_date == date(2004, 6, 17)
    assert record.amount == 100
    assert record.kid == '008000011688373'


def test_transaction_amount_item_1_for_avtalegiro_cancellation():
    record = netsgiro.records.TransactionAmountItem1.from_string(
        'NY2193300000001170604           00000000'
        '000000100          008000011688373000000'
    )

    assert record.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.TRANSACTION_AMOUNT_ITEM_1

    assert record.transaction_type == (
        netsgiro.TransactionType.AVTALEGIRO_CANCELLATION)
    assert record.transaction_number == 1

    assert record.nets_date == date(2004, 6, 17)
    assert record.amount == 100
    assert record.kid == '008000011688373'


def test_transaction_amount_item_1_for_ocr_giro_transactions():
    record = netsgiro.records.TransactionAmountItem1.from_string(
        'NY09103000000012001921320101464000000000'
        '000102000                  0000531000000'
    )

    assert record.service_code == netsgiro.ServiceCode.OCR_GIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.TRANSACTION_AMOUNT_ITEM_1

    assert record.transaction_type == (
        netsgiro.TransactionType.FROM_GIRO_DEBITED_ACCOUNT)
    assert record.transaction_number == 1

    assert record.nets_date == date(1992, 1, 20)

    assert record.centre_id == '13'
    assert record.day_code == 20
    assert record.partial_settlement_number == 1
    assert record.partial_settlement_serial_number == '01464'
    assert record.sign == '0'

    assert record.amount == 102000
    assert record.kid == '0000531'


def test_transaction_amount_item_2_for_avtalegiro_payment_request():
    record = netsgiro.records.TransactionAmountItem2.from_string(
        'NY2121310000001NAVN                     '
        '                                   00000'
    )

    assert record.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.TRANSACTION_AMOUNT_ITEM_2

    assert record.transaction_type == (
        netsgiro.TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION)
    assert record.transaction_number == 1

    assert record.payer_name == 'NAVN'
    assert record.reference is None


def test_transaction_amount_item_2_for_ocr_giro_transactions():
    record = netsgiro.records.TransactionAmountItem2.from_string(
        'NY09103100000019636827194099038562000000'
        '0160192999905123410000000000000000000000'
    )

    assert record.service_code == netsgiro.ServiceCode.OCR_GIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.TRANSACTION_AMOUNT_ITEM_2

    assert record.transaction_type == (
        netsgiro.TransactionType.FROM_GIRO_DEBITED_ACCOUNT)
    assert record.transaction_number == 1

    assert record.form_number == '9636827194'
    assert record.payer_name is None
    assert record.reference == '099038562'
    assert record.bank_date == date(1992, 1, 16)
    assert record.debit_account == '99990512341'


def test_transaction_amount_item_2_for_ocr_giro_with_data_in_filler_field():
    record = netsgiro.records.TransactionAmountItem2.from_string(
        'NY09103100000029797596016097596016188320'
        '6160192999910055240000000000000000000000'
    )

    assert record.service_code == netsgiro.ServiceCode.OCR_GIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.TRANSACTION_AMOUNT_ITEM_2

    assert record.transaction_type == (
        netsgiro.TransactionType.FROM_GIRO_DEBITED_ACCOUNT)
    assert record.transaction_number == 2

    assert record.form_number == '9797596016'
    assert record.payer_name is None
    assert record.reference == '097596016'
    assert record.bank_date == date(1992, 1, 16)
    assert record.debit_account == '99991005524'
    assert record._filler == '1883206'


def test_transaction_amount_item_3_for_ocr_giro_transactions():
    record = netsgiro.records.TransactionAmountItem3.from_string(
        'NY0921320000001Foo bar baz              '
        '               0000000000000000000000000'
    )

    assert record.service_code == netsgiro.ServiceCode.OCR_GIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.TRANSACTION_AMOUNT_ITEM_3

    assert record.transaction_type == (
        netsgiro.TransactionType.PURCHASE_WITH_TEXT)
    assert record.transaction_number == 1

    assert record.text == 'Foo bar baz'


def test_transaction_specification_for_avtalegiro_payment_request():
    record = netsgiro.records.TransactionSpecification.from_string(
        'NY212149000000140011 Gjelder Faktura: 16'
        '8837  Dato: 19/03/0400000000000000000000'
    )

    assert record.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.TRANSACTION_SPECIFICATION

    assert record.transaction_type == (
        netsgiro.TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION)
    assert record.transaction_number == 1

    assert record.line_number == 1
    assert record.column_number == 1
    assert record.text == ' Gjelder Faktura: 168837  Dato: 19/03/04'


def make_specification_records(num_lines, num_columns=2):
    return [
        netsgiro.records.TransactionSpecification(
            service_code=netsgiro.ServiceCode.AVTALEGIRO,
            transaction_type=(
                netsgiro.TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION),
            transaction_number=1,
            line_number=line,
            column_number=column,
            text='Line {}, column {}'.format(line, column),
        )
        for line in range(1, num_lines + 1)
        for column in range(1, num_columns + 1)
    ]


def test_transaction_specification_to_text_with_max_number_of_records():
    records = make_specification_records(42)

    result = netsgiro.records.TransactionSpecification.to_text(records)

    assert len(result.splitlines()) == 42
    assert 'Line 1, column 1' in result
    assert 'Line 42, column 2' in result


def test_transaction_specification_to_text_with_too_many_records():
    records = make_specification_records(43)

    with pytest.raises(ValueError) as exc_info:
        netsgiro.records.TransactionSpecification.to_text(records)

    assert 'Max 84 specification records allowed, got 86' in str(exc_info)


def test_avtalegiro_active_agreement():
    record = netsgiro.records.AvtaleGiroAgreement.from_string(
        'NY21947000000010          00800001168837'
        '3J00000000000000000000000000000000000000'
    )

    assert record.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.TRANSACTION_AGREEMENTS

    assert record.transaction_type == (
        netsgiro.TransactionType.AVTALEGIRO_AGREEMENT)
    assert record.transaction_number == 1

    assert record.registration_type == (
        netsgiro.AvtaleGiroRegistrationType.ACTIVE_AGREEMENT)
    assert record.kid == '008000011688373'
    assert record.notify is True


def test_avtalegiro_new_or_updated_agreement():
    record = netsgiro.records.AvtaleGiroAgreement.from_string(
        'NY21947000000011          00800001168837'
        '3N00000000000000000000000000000000000000'
    )

    assert record.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert record.RECORD_TYPE == netsgiro.RecordType.TRANSACTION_AGREEMENTS

    assert record.transaction_type == (
        netsgiro.TransactionType.AVTALEGIRO_AGREEMENT)
    assert record.transaction_number == 1

    assert record.registration_type == (
        netsgiro.AvtaleGiroRegistrationType.NEW_OR_UPDATED_AGREEMENT)
    assert record.kid == '008000011688373'
    assert record.notify is False



from datetime import date
from decimal import Decimal

import netsgiro


def test_parse_agreements(agreements_data):
    transmission = netsgiro.parse(agreements_data)

    assert isinstance(transmission, netsgiro.Transmission)
    assert transmission.number == '1091949'
    assert transmission.data_transmitter == netsgiro.NETS_ID
    assert transmission.data_recipient == '00010200'
    assert transmission.date == date(2017, 4, 19)
    assert len(transmission.assignments) == 1

    assignment = transmission.assignments[0]

    assert isinstance(assignment, netsgiro.Assignment)
    assert assignment.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert assignment.type == netsgiro.AssignmentType.AVTALEGIRO_AGREEMENTS
    assert assignment.agreement_id is None
    assert assignment.number == '0000002'
    assert assignment.account == '99991042764'
    assert len(assignment.transactions) == 16

    agreement_1 = assignment.transactions[0]

    assert isinstance(agreement_1, netsgiro.Agreement)
    assert agreement_1.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert agreement_1.TRANSACTION_TYPE == (
        netsgiro.TransactionType.AVTALEGIRO_AGREEMENT)
    assert agreement_1.number == 1

    assert agreement_1.registration_type == (
        netsgiro.AvtaleGiroRegistrationType.NEW_OR_UPDATED_AGREEMENT)
    assert agreement_1.kid == '000112000507155'
    assert agreement_1.notify is True

    agreement_2 = assignment.transactions[1]

    assert isinstance(agreement_2, netsgiro.Agreement)
    assert agreement_2.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert agreement_2.TRANSACTION_TYPE == (
        netsgiro.TransactionType.AVTALEGIRO_AGREEMENT)
    assert agreement_2.number == 2

    assert agreement_2.registration_type == (
        netsgiro.AvtaleGiroRegistrationType.NEW_OR_UPDATED_AGREEMENT)
    assert agreement_2.kid == '001006300507304'
    assert agreement_2.notify is False


def test_parse_payment_request(payment_request_data):
    transmission = netsgiro.parse(payment_request_data)

    assert isinstance(transmission, netsgiro.Transmission)
    assert transmission.number == '1000081'
    assert transmission.data_transmitter == '55555555'
    assert transmission.data_recipient == netsgiro.NETS_ID
    assert transmission.date == date(2004, 6, 17)
    assert len(transmission.assignments) == 1

    assignment = transmission.assignments[0]

    assert isinstance(assignment, netsgiro.Assignment)
    assert assignment.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert assignment.type == netsgiro.AssignmentType.TRANSACTIONS
    assert assignment.agreement_id == '000000000'
    assert assignment.number == '4000086'
    assert assignment.account == '88888888888'
    assert len(assignment.transactions) == 6

    transaction = assignment.transactions[0]

    assert isinstance(transaction, netsgiro.PaymentRequest)
    assert transaction.service_code == netsgiro.ServiceCode.AVTALEGIRO
    assert transaction.type == (
        netsgiro.TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION)
    assert transaction.number == 1
    assert transaction.date == date(2004, 6, 17)
    assert transaction.amount == Decimal('1.00')
    assert transaction.amount_in_cents == 100
    assert transaction.kid == '008000011688373'
    assert transaction.reference is None
    assert transaction.text == (
        ' Gjelder Faktura: 168837  Dato: 19/03/04'
        '                  ForfallsDato: 17/06/04\n'
    )

    # Specific to AvtaleGiro
    assert transaction.payer_name == 'NAVN'


def test_parse_ocr_giro_transactions(ocr_giro_transactions_data):
    transmission = netsgiro.parse(ocr_giro_transactions_data)

    assert isinstance(transmission, netsgiro.Transmission)
    assert transmission.number == '0170031'
    assert transmission.data_transmitter == netsgiro.NETS_ID
    assert transmission.data_recipient == '00010200'
    assert transmission.date == date(1992, 1, 20)
    assert len(transmission.assignments) == 1

    assignment = transmission.assignments[0]

    assert isinstance(assignment, netsgiro.Assignment)
    assert assignment.service_code == netsgiro.ServiceCode.OCR_GIRO
    assert assignment.type == netsgiro.AssignmentType.TRANSACTIONS
    assert assignment.agreement_id == '001008566'
    assert assignment.number == '0000002'
    assert assignment.account == '99991042764'
    assert len(assignment.transactions) == 20

    transaction = assignment.transactions[0]

    assert isinstance(transaction, netsgiro.Transaction)
    assert transaction.service_code == netsgiro.ServiceCode.OCR_GIRO
    assert transaction.type == (
        netsgiro.TransactionType.PURCHASE_WITH_TEXT)
    assert transaction.number == 1
    assert transaction.date == date(1992, 1, 20)
    assert transaction.amount == Decimal('1020')
    assert transaction.amount_in_cents == 102000
    assert transaction.kid == '0000531'
    assert transaction.reference == '099038562'
    assert transaction.text == 'Foo bar baz'

    # Specific to OCR Giro
    assert transaction.centre_id == '13'
    assert transaction.day_code == 20
    assert transaction.partial_settlement_number == 1
    assert transaction.partial_settlement_serial_number == '01464'
    assert transaction.sign == '0'
    assert transaction.form_number == '9636827194'
    assert transaction.bank_date == date(1992, 1, 16)
    assert transaction.debit_account == '99990512341'



from datetime import date

import pytest

import netsgiro
import netsgiro.records


def test_transmission_start():
    record = netsgiro.records.TransmissionStart(
        service_code=netsgiro.ServiceCode.NONE,
        transmission_number='1000081',
        data_transmitter='55555555',
        data_recipient=netsgiro.NETS_ID,
    )

    assert record.to_ocr() == (
        'NY00001055555555100008100008080000000000'
        '0000000000000000000000000000000000000000'
    )


@pytest.mark.parametrize(
    'transmission_number, data_transmitter, data_recipient, exc', [
        ('81', '55555555', '44444444', ValueError),
        (1000081, '55555555', '44444444', TypeError),
        ('1000081', '5555', '44444444', ValueError),
        ('1000081', 55555555, '44444444', TypeError),
        ('1000081', '55555555', '4444', ValueError),
        ('1000081', '55555555', 44444444, TypeError),
    ])
def test_transmission_start_with_invalid_data(
        transmission_number, data_transmitter, data_recipient, exc):
    with pytest.raises(exc):
        netsgiro.records.TransmissionStart(
            service_code=netsgiro.ServiceCode.NONE,
            transmission_number=transmission_number,
            data_transmitter=data_transmitter,
            data_recipient=data_recipient,
        )


def test_transmission_end():
    record = netsgiro.records.TransmissionEnd(
        service_code=netsgiro.ServiceCode.NONE,
        num_transactions=6,
        num_records=22,
        total_amount=600,
        nets_date=date(2004, 6, 17),
    )

    assert record.to_ocr() == (
        'NY00008900000006000000220000000000000060'
        '0170604000000000000000000000000000000000'
    )


@pytest.mark.parametrize(
    'num_tx, num_records, total_amount, nets_date, exc', [
        ('abc', 22, 600, date(2004, 6, 17), ValueError),
        (6, 'abc', 600, date(2004, 6, 17), ValueError),
        (6, 22, 'abc', date(2004, 6, 17), ValueError),
        (6, 22, 600, '2004-06-17', ValueError),
    ])
def test_transmission_end_with_invalid_data(
        num_tx, num_records, total_amount, nets_date, exc):
    with pytest.raises(exc):
        netsgiro.records.TransmissionEnd(
            service_code=netsgiro.ServiceCode.NONE,
            num_transactions=num_tx,
            num_records=num_records,
            total_amount=total_amount,
            nets_date=nets_date,
        )


def test_assignment_start_for_avtalegiro_payment_requests():
    record = netsgiro.records.AssignmentStart(
        service_code=netsgiro.ServiceCode.AVTALEGIRO,
        assignment_type=netsgiro.AssignmentType.TRANSACTIONS,
        assignment_number='4000086',
        assignment_account='88888888888',
        agreement_id='000000111',
    )

    assert record.to_ocr() == (
        'NY21002000000011140000868888888888800000'
        '0000000000000000000000000000000000000000'
    )


def test_assignment_start_for_avtalegiro_agreements():
    record = netsgiro.records.AssignmentStart(
        service_code=netsgiro.ServiceCode.AVTALEGIRO,
        assignment_type=netsgiro.AssignmentType.AVTALEGIRO_AGREEMENTS,
        assignment_number='4000086',
        assignment_account='88888888888',
        agreement_id=None,
    )

    assert record.to_ocr() == (
        'NY21242000000000040000868888888888800000'
        '0000000000000000000000000000000000000000'
    )


@pytest.mark.parametrize(
    'number, account, agreement_id, exc', [
        (4000086, '88888888888', '123456789', TypeError),
        ('86', '88888888888', '123456789', ValueError),
        ('4000086', 88888888888, '123456789', TypeError),
        ('4000086', '88', '123456789', ValueError),
        ('4000086', '88888888888', 123456789, TypeError),
        ('4000086', '88888888888', '6789', ValueError),
    ])
def test_assignment_start_with_invalid_data(
        number, account, agreement_id, exc):
    with pytest.raises(exc):
        netsgiro.records.AssignmentStart(
            service_code=netsgiro.ServiceCode.AVTALEGIRO,
            assignment_type=netsgiro.AssignmentType.TRANSACTIONS,
            assignment_number=number,
            assignment_account=account,
            agreement_id=agreement_id,
        )


def test_assignment_end_for_avtalegiro_payment_requests():
    record = netsgiro.records.AssignmentEnd(
        service_code=netsgiro.ServiceCode.AVTALEGIRO,
        assignment_type=netsgiro.AssignmentType.TRANSACTIONS,
        num_transactions=6,
        num_records=20,
        total_amount=600,
        nets_date_1=date(2004, 6, 17),
        nets_date_2=date(2004, 6, 17),
    )

    assert record.to_ocr() == (
        'NY21008800000006000000200000000000000060'
        '0170604170604000000000000000000000000000'
    )


def test_assignment_end_for_avtalegiro_agreements():
    record = netsgiro.records.AssignmentEnd(
        service_code=netsgiro.ServiceCode.AVTALEGIRO,
        assignment_type=netsgiro.AssignmentType.AVTALEGIRO_AGREEMENTS,
        num_transactions=6,
        num_records=20,
    )

    assert record.to_ocr() == (
        'NY21248800000006000000200000000000000000'
        '0000000000000000000000000000000000000000'
    )


def test_transaction_amount_item_1_for_ocr_giro_transactions():
    record = netsgiro.records.TransactionAmountItem1(
        service_code=netsgiro.ServiceCode.OCR_GIRO,
        transaction_type=netsgiro.TransactionType.FROM_GIRO_DEBITED_ACCOUNT,
        transaction_number=1,
        nets_date=date(1992, 1, 20),
        amount=102000,
        kid='0000531',

        centre_id='13',
        day_code=20,
        partial_settlement_number=1,
        partial_settlement_serial_number='01464',
        sign='0',
    )

    assert record.to_ocr() == (
        'NY09103000000012001921320101464000000000'
        '000102000                  0000531000000'
    )


def test_transaction_amount_item_1_for_avtalegiro_payment_requests():
    record = netsgiro.records.TransactionAmountItem1(
        service_code=netsgiro.ServiceCode.AVTALEGIRO,
        transaction_type=(
            netsgiro.TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION),
        transaction_number=1,
        nets_date=date(2004, 6, 17),
        amount=100,
        kid='008000011688373',
    )

    assert record.to_ocr() == (
        'NY2121300000001170604           00000000'
        '000000100          008000011688373000000'
    )


def test_transaction_amount_item_2_for_avtalegiro_payment_request():
    record = netsgiro.records.TransactionAmountItem2(
        service_code=netsgiro.ServiceCode.AVTALEGIRO,
        transaction_type=(
            netsgiro.TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION),
        transaction_number=1,
        reference=None,

        payer_name='NAVN',
    )

    assert record.to_ocr() == (
        'NY2121310000001NAVN                     '
        '                                   00000'
    )


def test_transaction_amount_item_2_cuts_too_long_payer_name():
    record = netsgiro.records.TransactionAmountItem2(
        service_code=netsgiro.ServiceCode.AVTALEGIRO,
        transaction_type=(
            netsgiro.TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION),
        transaction_number=1,
        reference=None,

        payer_name='NAVN123456789',
    )

    assert record.to_ocr() == (
        'NY2121310000001NAVN123456               '
        '                                   00000'
    )


def test_transaction_amount_item_2_for_ocr_giro_transactions():
    record = netsgiro.records.TransactionAmountItem2(
        service_code=netsgiro.ServiceCode.OCR_GIRO,
        transaction_type=(
            netsgiro.TransactionType.FROM_GIRO_DEBITED_ACCOUNT),
        transaction_number=1,
        reference='099038562',

        form_number='9636827194',
        bank_date=date(1992, 1, 16),
        debit_account='99990512341',
    )

    assert record.to_ocr() == (
        'NY09103100000019636827194099038562000000'
        '0160192999905123410000000000000000000000'
    )


def test_transaction_amount_item_2_for_ocr_giro_without_bank_date():
    record = netsgiro.records.TransactionAmountItem2(
        service_code=netsgiro.ServiceCode.OCR_GIRO,
        transaction_type=(
            netsgiro.TransactionType.FROM_GIRO_DEBITED_ACCOUNT),
        transaction_number=1,
        reference='099038562',

        form_number='9636827194',
        bank_date=None,
        debit_account='99990512341',
    )

    assert record.to_ocr() == (
        'NY09103100000019636827194099038562000000'
        '0000000999905123410000000000000000000000'
    )


def test_transaction_amount_item_3_for_ocr_giro_transactions():
    record = netsgiro.records.TransactionAmountItem3(
        service_code=netsgiro.ServiceCode.OCR_GIRO,
        transaction_type=(
            netsgiro.TransactionType.PURCHASE_WITH_TEXT),
        transaction_number=1,
        text='Foo bar baz',
    )

    assert record.to_ocr() == (
        'NY0921320000001Foo bar baz              '
        '               0000000000000000000000000'
    )


def test_transaction_specification_for_avtalegiro_payment_request():
    record = netsgiro.records.TransactionSpecification(
        service_code=netsgiro.ServiceCode.AVTALEGIRO,
        transaction_type=(
            netsgiro.TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION),
        transaction_number=1,
        line_number=1,
        column_number=1,
        text=' Gjelder Faktura: 168837  Dato: 19/03/04',
    )

    assert record.to_ocr() == (
        'NY212149000000140011 Gjelder Faktura: 16'
        '8837  Dato: 19/03/0400000000000000000000'
    )


def test_transaction_specification_from_longer_text():
    records = list(netsgiro.records.TransactionSpecification.from_text(
        service_code=netsgiro.ServiceCode.AVTALEGIRO,
        transaction_type=(
            netsgiro.TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION),
        transaction_number=1,
        text=' Gjelder Faktura: 168837  Dato: 19/03/04\nFoo bar baz quux',
    ))

    assert len(records) == 4
    assert records[0].to_ocr() == (
        'NY212149000000140011 Gjelder Faktura: 16'
        '8837  Dato: 19/03/0400000000000000000000'
    )
    assert records[1].to_ocr() == (
        'NY212149000000140012                    '
        '                    00000000000000000000'
    )
    assert records[2].to_ocr() == (
        'NY212149000000140021Foo bar baz quux    '
        '                    00000000000000000000'
    )
    assert records[3].to_ocr() == (
        'NY212149000000140022                    '
        '                    00000000000000000000'
    )


def test_avtalegiro_agreement():
    record = netsgiro.records.AvtaleGiroAgreement(
        service_code=netsgiro.ServiceCode.AVTALEGIRO,
        transaction_type=(
            netsgiro.TransactionType.AVTALEGIRO_AGREEMENT),
        transaction_number=1,
        registration_type=(
            netsgiro.AvtaleGiroRegistrationType.NEW_OR_UPDATED_AGREEMENT),
        kid='008000011688373',
        notify=False,
    )

    assert record.to_ocr() == (
        'NY21947000000011          00800001168837'
        '3N00000000000000000000000000000000000000'
    )



"""The higher-level objects API."""

import collections
import datetime
from decimal import Decimal
from typing import Iterable, List, Mapping, Optional, Union

import attr
from attr.validators import instance_of, optional

import netsgiro
import netsgiro.records
from netsgiro.records import Record
from netsgiro.validators import str_of_length


__all__ = [
    'Transmission',
    'Assignment',
    'Agreement',
    'PaymentRequest',
    'Transaction',
    'parse',
]


@attr.s
class Transmission:
    """Transmission is the top-level object.

    An OCR file contains a single transmission. The transmission can contain
    multiple :class:`~netsgiro.Assignment` objects of various types.
    """

    #: Data transmitters unique enumeration of the transmission. String of 7
    #: digits.
    number = attr.ib(validator=str_of_length(7))

    #: Data transmitter's Nets ID. String of 8 digits.
    data_transmitter = attr.ib(validator=str_of_length(8))

    #: Data recipient's Nets ID. String of 8 digits.
    data_recipient = attr.ib(validator=str_of_length(8))

    #: For OCR Giro files from Nets, this is Nets' processing date.
    #:
    #: For AvtaleGiro payment request, the earliest due date in the
    #: transmission is automatically used.
    date = attr.ib(
        default=None, validator=optional(instance_of(datetime.date)))

    #: List of assignments.
    assignments = attr.ib(default=attr.Factory(list), repr=False)

    @classmethod
    def from_records(cls, records: List[Record]) -> 'Transmission':
        """Build a Transmission object from a list of record objects."""
        if len(records) < 2:
            raise ValueError(
                'At least 2 records required, got {}'.format(len(records)))

        start, body, end = records[0], records[1:-1], records[-1]

        assert isinstance(start, netsgiro.records.TransmissionStart)
        assert isinstance(end, netsgiro.records.TransmissionEnd)

        return cls(
            number=start.transmission_number,
            data_transmitter=start.data_transmitter,
            data_recipient=start.data_recipient,
            date=end.nets_date,
            assignments=cls._get_assignments(body),
        )

    @staticmethod
    def _get_assignments(records: List[Record]) -> List['Assignment']:
        assignments = collections.OrderedDict()

        current_assignment_number = None
        for record in records:
            if isinstance(record, netsgiro.records.AssignmentStart):
                current_assignment_number = record.assignment_number
                assignments[current_assignment_number] = []
            if current_assignment_number is None:
                raise ValueError(
                    'Expected AssignmentStart record, got {!r}'.format(record))
            assignments[current_assignment_number].append(record)
            if isinstance(record, netsgiro.records.AssignmentEnd):
                current_assignment_number = None

        return [Assignment.from_records(rs) for rs in assignments.values()]

    def to_ocr(self) -> str:
        """Convert the transmission to an OCR string."""
        lines = [record.to_ocr() for record in self.to_records()]
        return '\n'.join(lines)

    def to_records(self) -> Iterable[Record]:
        """Convert the transmission to a list of records."""
        yield self._get_start_record()
        for assignment in self.assignments:
            yield from assignment.to_records()
        yield self._get_end_record()

    def _get_start_record(self) -> Record:
        return netsgiro.records.TransmissionStart(
            service_code=netsgiro.ServiceCode.NONE,
            transmission_number=self.number,
            data_transmitter=self.data_transmitter,
            data_recipient=self.data_recipient,
        )

    def _get_end_record(self) -> Record:
        avtalegiro_payment_request = all(
            assignment.service_code == netsgiro.ServiceCode.AVTALEGIRO and
            assignment.type in (
                netsgiro.AssignmentType.TRANSACTIONS,
                netsgiro.AssignmentType.AVTALEGIRO_CANCELLATIONS,
            )
            for assignment in self.assignments)
        if self.assignments and avtalegiro_payment_request:
            date = min(
                assignment.get_earliest_transaction_date()
                for assignment in self.assignments)
        else:
            date = self.date

        return netsgiro.records.TransmissionEnd(
            service_code=netsgiro.ServiceCode.NONE,
            num_transactions=self.get_num_transactions(),
            num_records=self.get_num_records(),
            total_amount=int(self.get_total_amount() * 100),
            nets_date=date,
        )

    def add_assignment(
            self, *,
            service_code: netsgiro.ServiceCode,
            assignment_type: netsgiro.AssignmentType,
            agreement_id: Optional[str]=None,
            number: str,
            account: str,
            date: Optional[datetime.date]=None
            ) -> 'Assignment':
        """Add an assignment to the tranmission."""

        assignment = Assignment(
            service_code=service_code,
            type=assignment_type,
            agreement_id=agreement_id,
            number=number,
            account=account,
            date=date,
        )
        self.assignments.append(assignment)
        return assignment

    def get_num_transactions(self) -> int:
        """Get number of transactions in the transmission."""
        return sum(
            assignment.get_num_transactions()
            for assignment in self.assignments)

    def get_num_records(self) -> int:
        """Get number of records in the transmission.

        Includes the transmission's start and end record.
        """
        return 2 + sum(
            assignment.get_num_records()
            for assignment in self.assignments)

    def get_total_amount(self) -> Decimal:
        """Get the total amount from all transactions in the transmission."""
        return sum(
            assignment.get_total_amount()
            for assignment in self.assignments)


@attr.s
class Assignment:
    """An Assignment groups multiple transactions within a transmission.

    Use :meth:`netsgiro.Transmission.add_assignment` to create assignments.
    """

    #: The service code. One of :class:`~netsgiro.ServiceCode`.
    service_code = attr.ib(convert=netsgiro.ServiceCode)

    #: The transaction type. One of :class:`~netsgiro.TransactionType`.
    type = attr.ib(convert=netsgiro.AssignmentType)

    #: The assignment number. String of 7 digits.
    number = attr.ib(validator=str_of_length(7))

    #: The payee's bank account. String of 11 digits.
    account = attr.ib(validator=str_of_length(11))

    #: Used for OCR Giro.
    #:
    #: The payee's agreement ID with Nets. String of 9 digits.
    agreement_id = attr.ib(default=None, validator=optional(str_of_length(9)))

    #: Used for OCR Giro.
    #:
    #: The date the assignment was generated by Nets.
    date = attr.ib(
        default=None, validator=optional(instance_of(datetime.date)))

    #: List of transaction objects, like :class:`~netsgiro.Agreement`,
    #: :class:`~netsgiro.PaymentRequest`, :class:`~netsgiro.Transaction`.
    transactions = attr.ib(default=attr.Factory(list), repr=False)

    _next_transaction_number = 1

    @classmethod
    def from_records(cls, records: List[Record]) -> 'Assignment':
        """Build an Assignment object from a list of record objects."""
        if len(records) < 2:
            raise ValueError(
                'At least 2 records required, got {}'.format(len(records)))

        start, body, end = records[0], records[1:-1], records[-1]

        assert isinstance(start, netsgiro.records.AssignmentStart)
        assert isinstance(end, netsgiro.records.AssignmentEnd)

        if start.service_code == netsgiro.ServiceCode.AVTALEGIRO:
            if (
                    start.assignment_type ==
                    netsgiro.AssignmentType.AVTALEGIRO_AGREEMENTS):
                transactions = cls._get_agreements(body)
            else:
                transactions = cls._get_payment_requests(body)
        elif start.service_code == netsgiro.ServiceCode.OCR_GIRO:
            transactions = cls._get_transactions(body)
        else:
            raise ValueError(
                'Unknown service code: {}'.format(start.service_code))

        return cls(
            service_code=start.service_code,
            type=start.assignment_type,
            agreement_id=start.agreement_id,
            number=start.assignment_number,
            account=start.assignment_account,
            date=end.nets_date,
            transactions=transactions,
        )

    @staticmethod
    def _get_agreements(records: List[Record]) -> List['Agreement']:
        return [Agreement.from_records([r]) for r in records]

    @classmethod
    def _get_payment_requests(
            cls, records: List[Record]) -> List['PaymentRequest']:
        transactions = cls._group_by_transaction_number(records)
        return [
            PaymentRequest.from_records(rs) for rs in transactions.values()]

    @classmethod
    def _get_transactions(cls, records: List[Record]) -> List['Transaction']:
        transactions = cls._group_by_transaction_number(records)
        return [Transaction.from_records(rs) for rs in transactions.values()]

    @staticmethod
    def _group_by_transaction_number(
            records: List[Record]) -> Mapping[int, List[Record]]:
        transactions = collections.OrderedDict()

        for record in records:
            if record.transaction_number not in transactions:
                transactions[record.transaction_number] = []
            transactions[record.transaction_number].append(record)

        return transactions

    def to_records(self) -> Iterable[Record]:
        """Convert the assignment to a list of records."""
        yield self._get_start_record()
        for transaction in self.transactions:
            yield from transaction.to_records()
        yield self._get_end_record()

    def _get_start_record(self) -> Record:
        return netsgiro.records.AssignmentStart(
            service_code=self.service_code,
            assignment_type=self.type,
            assignment_number=self.number,
            assignment_account=self.account,
            agreement_id=self.agreement_id,
        )

    def _get_end_record(self) -> Record:
        if self.service_code == netsgiro.ServiceCode.OCR_GIRO:
            dates = {
                'nets_date_1': self.date,
                'nets_date_2': self.get_earliest_transaction_date(),
                'nets_date_3': self.get_latest_transaction_date(),
            }
        elif self.service_code == netsgiro.ServiceCode.AVTALEGIRO:
            dates = {
                'nets_date_1': self.get_earliest_transaction_date(),
                'nets_date_2': self.get_latest_transaction_date(),
            }
        else:
            raise ValueError(
                'Unhandled service code: {}'.format(self.service_code))

        return netsgiro.records.AssignmentEnd(
            service_code=self.service_code,
            assignment_type=self.type,
            num_transactions=self.get_num_transactions(),
            num_records=self.get_num_records(),
            total_amount=int(self.get_total_amount() * 100),
            **dates
        )

    def add_payment_request(
            self, *,
            kid: str,
            due_date: datetime.date,
            amount: Decimal,
            reference: Optional[str]=None,
            payer_name: Optional[str]=None,
            bank_notification: Union[bool, str]=False
            ) -> 'Transaction':
        """Add an AvtaleGiro payment request to the assignment.

        The assignment must have service code
        :attr:`~netsgiro.ServiceCode.AVTALEGIRO` and assignment type
        :attr:`~netsgiro.AssignmentType.TRANSACTIONS`.
        """

        assert self.service_code == netsgiro.ServiceCode.AVTALEGIRO, (
            'Can only add payment requests to AvtaleGiro assignments')
        assert self.type == netsgiro.AssignmentType.TRANSACTIONS, (
            'Can only add payment requests to transaction assignments')

        if bank_notification:
            transaction_type = (
                netsgiro.TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION)
        else:
            transaction_type = (
                netsgiro.TransactionType.AVTALEGIRO_WITH_PAYEE_NOTIFICATION)

        return self._add_avtalegiro_transaction(
            transaction_type=transaction_type,
            kid=kid,
            due_date=due_date,
            amount=amount,
            reference=reference,
            payer_name=payer_name,
            bank_notification=bank_notification,
        )

    def add_payment_cancellation(
            self, *,
            kid: str,
            due_date: datetime.date,
            amount: Decimal,
            reference: Optional[str]=None,
            payer_name: Optional[str]=None,
            bank_notification: Union[bool, str]=False
            ) -> 'Transaction':
        """Add an AvtaleGiro cancellation to the assignment.

        The assignment must have service code
        :attr:`~netsgiro.ServiceCode.AVTALEGIRO` and assignment type
        :attr:`~netsgiro.AssignmentType.AVTALEGIRO_CANCELLATIONS`.

        Otherwise, the cancellation must be identical to the payment request it
        is cancelling.
        """

        assert self.service_code == netsgiro.ServiceCode.AVTALEGIRO, (
            'Can only add cancellation to AvtaleGiro assignments')
        assert self.type == netsgiro.AssignmentType.AVTALEGIRO_CANCELLATIONS, (
            'Can only add cancellation to cancellation assignments')

        return self._add_avtalegiro_transaction(
            transaction_type=netsgiro.TransactionType.AVTALEGIRO_CANCELLATION,
            kid=kid,
            due_date=due_date,
            amount=amount,
            reference=reference,
            payer_name=payer_name,
            bank_notification=bank_notification,
        )

    def _add_avtalegiro_transaction(
            self, *,
            transaction_type,
            kid, due_date, amount,
            reference=None, payer_name=None, bank_notification=None
            ) -> 'Transaction':

        if isinstance(bank_notification, str):
            text = bank_notification
        else:
            text = ''

        number = self._next_transaction_number
        self._next_transaction_number += 1

        transaction = PaymentRequest(
            service_code=self.service_code,
            type=transaction_type,
            number=number,
            date=due_date,
            amount=amount,
            kid=kid,
            reference=reference,
            text=text,

            payer_name=payer_name,
        )
        self.transactions.append(transaction)
        return transaction

    def get_num_transactions(self) -> int:
        """Get number of transactions in the assignment."""
        return len(self.transactions)

    def get_num_records(self) -> int:
        """Get number of records in the assignment.

        Includes the assignment's start and end record.
        """

        return 2 + sum(
            len(list(transaction.to_records()))
            for transaction in self.transactions)

    def get_total_amount(self) -> Decimal:
        """Get the total amount from all transactions in the assignment."""
        transactions = [
            transaction
            for transaction in self.transactions
            if hasattr(transaction, 'amount')]
        if not transactions:
            return Decimal(0)
        return sum(transaction.amount for transaction in transactions)

    def get_earliest_transaction_date(self) -> Optional[datetime.date]:
        """Get earliest date from the assignment's transactions."""
        transactions = [
            transaction
            for transaction in self.transactions
            if hasattr(transaction, 'date')]
        if not transactions:
            return None
        return min(transaction.date for transaction in transactions)

    def get_latest_transaction_date(self) -> Optional[datetime.date]:
        """Get latest date from the assignment's transactions."""
        transactions = [
            transaction
            for transaction in self.transactions
            if hasattr(transaction, 'date')]
        if not transactions:
            return None
        return max(transaction.date for transaction in transactions)


@attr.s
class Agreement:
    """Agreement contains an AvtaleGiro agreement update.

    Agreements are only found in assignments of the
    :attr:`~netsgiro.AssignmentType.AVTALEGIRO_AGREEMENTS` type, which are only
    created by Nets.
    """

    #: The service code. One of :class:`~netsgiro.ServiceCode`.
    service_code = attr.ib(convert=netsgiro.ServiceCode)

    #: Transaction number. Unique and ordered within an assignment.
    number = attr.ib(validator=instance_of(int))

    #: Type of agreement registration update.
    #: One of :class:`~netsgiro.AvtaleGiroRegistrationType`.
    registration_type = attr.ib(convert=netsgiro.AvtaleGiroRegistrationType)

    #: KID number to identify the customer and invoice.
    kid = attr.ib(validator=optional(instance_of(str)))

    #: Whether the payer wants notification about payment requests.
    notify = attr.ib(validator=instance_of(bool))

    TRANSACTION_TYPE = netsgiro.TransactionType.AVTALEGIRO_AGREEMENT

    @classmethod
    def from_records(cls, records: List[Record]) -> 'Agreement':
        """Build an Agreement object from a list of record objects."""

        assert len(records) == 1
        record = records[0]
        assert isinstance(record, netsgiro.records.AvtaleGiroAgreement)
        assert (
            record.transaction_type ==
            netsgiro.TransactionType.AVTALEGIRO_AGREEMENT)

        return cls(
            service_code=record.service_code,
            number=record.transaction_number,

            registration_type=record.registration_type,
            kid=record.kid,
            notify=record.notify,
        )

    def to_records(self) -> Iterable[Record]:
        """Convert the agreement to a list of records."""
        yield netsgiro.records.AvtaleGiroAgreement(
            service_code=self.service_code,
            transaction_type=self.TRANSACTION_TYPE,
            transaction_number=self.number,

            registration_type=self.registration_type,
            kid=self.kid,
            notify=self.notify,
        )


@attr.s
class PaymentRequest:
    """PaymentRequest contains an AvtaleGiro payment request or cancellation.

    To create a transaction, you will normally use the helper methods on
    :class:`~netsgiro.Assignment`:
    :meth:`~netsgiro.Assignment.add_payment_request` and
    :meth:`~netsgiro.Assignment.add_payment_cancellation`.
    """

    #: The service code. One of :class:`~netsgiro.ServiceCode`.
    service_code = attr.ib(convert=netsgiro.ServiceCode)

    #: The transaction type. One of :class:`~netsgiro.TransactionType`.
    type = attr.ib(convert=netsgiro.TransactionType)

    #: Transaction number. Unique and ordered within an assignment.
    number = attr.ib(validator=instance_of(int))

    #: The due date.
    date = attr.ib(validator=instance_of(datetime.date))

    #: Transaction amount in NOK with two decimals.
    amount = attr.ib(convert=Decimal)

    #: KID number to identify the customer and invoice.
    kid = attr.ib(validator=optional(instance_of(str)))

    #: This is a specification line that will, if set, be displayed on the
    #: payers account statement. Alphanumeric, max 25 chars.
    reference = attr.ib(validator=optional(instance_of(str)))

    #: This is up to 42 lines of 80 chars each of free text used by the bank to
    #: notify the payer about the payment request. It is not used if the payee
    #: is responsible for notifying the payer.
    text = attr.ib(validator=optional(instance_of(str)))

    #: The value is only used to help the payee cross-reference reports from
    #: Nets with their own records. It is not visible to the payer.
    payer_name = attr.ib(validator=optional(instance_of(str)))

    @property
    def amount_in_cents(self) -> int:
        """Transaction amount in NOK cents."""
        return int(self.amount * 100)

    @classmethod
    def from_records(cls, records: List[Record]) -> 'Transaction':
        """Build a Transaction object from a list of record objects."""
        amount_item_1 = records.pop(0)
        assert isinstance(
            amount_item_1, netsgiro.records.TransactionAmountItem1)
        amount_item_2 = records.pop(0)
        assert isinstance(
            amount_item_2, netsgiro.records.TransactionAmountItem2)

        text = netsgiro.records.TransactionSpecification.to_text(records)

        return cls(
            service_code=amount_item_1.service_code,
            type=amount_item_1.transaction_type,
            number=amount_item_1.transaction_number,
            date=amount_item_1.nets_date,
            amount=Decimal(amount_item_1.amount) / 100,
            kid=amount_item_1.kid,
            reference=amount_item_2.reference,
            text=text,

            payer_name=amount_item_2.payer_name,
        )

    def to_records(self) -> Iterable[Record]:
        """Convert the transaction to a list of records."""
        yield netsgiro.records.TransactionAmountItem1(
            service_code=self.service_code,
            transaction_type=self.type,
            transaction_number=self.number,

            nets_date=self.date,
            amount=self.amount_in_cents,
            kid=self.kid,
        )
        yield netsgiro.records.TransactionAmountItem2(
            service_code=self.service_code,
            transaction_type=self.type,
            transaction_number=self.number,

            reference=self.reference,
            payer_name=self.payer_name,
        )

        if self.type == (
                netsgiro.TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION):
            yield from netsgiro.records.TransactionSpecification.from_text(
                service_code=self.service_code,
                transaction_type=self.type,
                transaction_number=self.number,

                text=self.text,
            )


@attr.s
class Transaction:
    """Transaction contains an OCR Giro transaction.

    Transactions are found in assignments with the service code
    :attr:`~netsgiro.ServiceCode.OCR_GIRO` type, which are only
    created by Nets.
    """

    #: The service code. One of :class:`~netsgiro.ServiceCode`.
    service_code = attr.ib(convert=netsgiro.ServiceCode)

    #: The transaction type. One of :class:`~netsgiro.TransactionType`.
    type = attr.ib(convert=netsgiro.TransactionType)

    #: Transaction number. Unique and ordered within an assignment.
    number = attr.ib(validator=instance_of(int))

    #: Nets' processing date.
    date = attr.ib(validator=instance_of(datetime.date))

    #: Transaction amount in NOK with two decimals.
    amount = attr.ib(convert=Decimal)

    #: KID number to identify the customer and invoice.
    kid = attr.ib(validator=optional(instance_of(str)))

    #: The value depends on the payment method.
    reference = attr.ib(validator=optional(instance_of(str)))

    #: Up to 40 chars of free text from the payment terminal.
    text = attr.ib(validator=optional(instance_of(str)))

    #: Used for OCR Giro.
    centre_id = attr.ib(validator=optional(str_of_length(2)))

    #: Used for OCR Giro.
    day_code = attr.ib(validator=optional(instance_of(int)))

    #: Used for OCR Giro.
    partial_settlement_number = attr.ib(validator=optional(instance_of(int)))

    #: Used for OCR Giro.
    partial_settlement_serial_number = attr.ib(
        validator=optional(str_of_length(5)))

    #: Used for OCR Giro.
    sign = attr.ib(validator=optional(str_of_length(1)))

    #: Used for OCR Giro.
    form_number = attr.ib(validator=optional(str_of_length(10)))

    #: Used for OCR Giro.
    bank_date = attr.ib(validator=optional(instance_of(datetime.date)))

    #: Used for OCR Giro.
    debit_account = attr.ib(validator=optional(str_of_length(11)))

    _filler = attr.ib(validator=optional(str_of_length(7)))

    @property
    def amount_in_cents(self) -> int:
        """Transaction amount in NOK cents."""
        return int(self.amount * 100)

    @classmethod
    def from_records(cls, records: List[Record]) -> 'Transaction':
        """Build a Transaction object from a list of record objects."""
        amount_item_1 = records.pop(0)
        assert isinstance(
            amount_item_1, netsgiro.records.TransactionAmountItem1)
        amount_item_2 = records.pop(0)
        assert isinstance(
            amount_item_2, netsgiro.records.TransactionAmountItem2)

        if (
                len(records) == 1 and
                isinstance(
                    records[0], netsgiro.records.TransactionAmountItem3)):
            text = records[0].text
        else:
            text = None

        return cls(
            service_code=amount_item_1.service_code,
            type=amount_item_1.transaction_type,
            number=amount_item_1.transaction_number,
            date=amount_item_1.nets_date,
            amount=Decimal(amount_item_1.amount) / 100,
            kid=amount_item_1.kid,
            reference=amount_item_2.reference,
            text=text,

            centre_id=amount_item_1.centre_id,
            day_code=amount_item_1.day_code,
            partial_settlement_number=amount_item_1.partial_settlement_number,
            partial_settlement_serial_number=(
                amount_item_1.partial_settlement_serial_number),
            sign=amount_item_1.sign,
            form_number=amount_item_2.form_number,
            bank_date=amount_item_2.bank_date,
            debit_account=amount_item_2.debit_account,
            filler=amount_item_2._filler,
        )

    def to_records(self) -> Iterable[Record]:
        """Convert the transaction to a list of records."""
        yield netsgiro.records.TransactionAmountItem1(
            service_code=self.service_code,
            transaction_type=self.type,
            transaction_number=self.number,

            nets_date=self.date,
            amount=self.amount_in_cents,
            kid=self.kid,

            centre_id=self.centre_id,
            day_code=self.day_code,
            partial_settlement_number=self.partial_settlement_number,
            partial_settlement_serial_number=(
                self.partial_settlement_serial_number),
            sign=self.sign,
        )
        yield netsgiro.records.TransactionAmountItem2(
            service_code=self.service_code,
            transaction_type=self.type,
            transaction_number=self.number,

            reference=self.reference,
            form_number=self.form_number,
            bank_date=self.bank_date,
            debit_account=self.debit_account,
            filler=self._filler,
        )

        if self.type in (
                netsgiro.TransactionType.REVERSING_WITH_TEXT,
                netsgiro.TransactionType.PURCHASE_WITH_TEXT):
            yield netsgiro.records.TransactionAmountItem3(
                service_code=self.service_code,
                transaction_type=self.type,
                transaction_number=self.number,

                text=self.text,
            )


def parse(data: str) -> Transmission:
    """Parse an OCR file into a Transmission object."""
    records = netsgiro.records.parse(data)
    return Transmission.from_records(records)



import re

from setuptools import find_packages, setup


with open('netsgiro/__init__.py') as fh:
    metadata = dict(re.findall("__([a-z]+)__ = '([^']+)'", fh.read()))


with open('README.rst') as fh:
    long_description = fh.read()


setup(
    name='netsgiro',
    version=metadata['version'],
    description='File parsers for Nets AvtaleGiro and OCR Giro',
    long_description=long_description,
    url='https://netsgiro.readthedocs.io/',
    author='Otovo AS',
    author_email='jodal+netsgiro@otovo.no',
    license='Apache License, Version 2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='avtalegiro ocr giro',
    packages=find_packages(exclude=['tests', 'tests.*']),
    install_requires=[
        'attrs',
        'typing',  # Needed for Python 3.4
    ],
    extras_require={
        'dev': [
            'check-manifest',
            'flake8',
            'flake8-import-order',
            'mypy',
            'pydocstyle',
            'pytest',
            'pytest-xdist',
            'sphinx',
            'sphinx_rtd_theme',
            'tox',
        ],
    },
)
